# pcb-rnd user script: tent or untent and/or paste or unpaste vias
# Copyright (C) 2019 Tibor 'Igor2' Palinkas
# licensed under the GNU GPL verion 2 or later.

BEGIN {
# action target layers
	top_mask = 0;     FIDNAME[top_mask] = "top_mask"; FIDSRC[top_mask] = "top_copper";
	bottom_mask = 1;  FIDNAME[bottom_mask] = "bottom_mask"; FIDSRC[bottom_mask] = "bottom_copper";
	top_paste = 2;    FIDNAME[top_paste] = "top_paste"; FIDSRC[top_paste] = "top_copper";
	bottom_paste = 3; FIDNAME[bottom_paste] = "bottom_paste"; FIDSRC[bottom_paste] = "bottom_copper";
	fids = 4

# actions
	add = +1
	remove = -1
}

# Execute all actions in ACT[action_target] on all vias; if filt is
# specified, also use that expression for the search with &&. If noterm
# is +1, operate only on non-terminal padstacks ("vias"), if -1 perate only
# on terminals, if 0 operate on any padstack.
function via_tent_(ACT, noterm,   filt,   a,FID,list,acts,needs_action,pid,proto)
{

# sanity check on actions and count non-nops
	for(a = 0; a < fids; a++) {
		ACT[a] = int(ACT[a])
		if (ACT[a] != 0)
			acts++
	}

# if there is no action to perform, just return
	if (acts == 0)
		return 0

# precompile shape lookups for speeding up the search
	for(a = 0; a < fids; a++)
		FID[a] = querycompilefield("compile", ".shape." FIDNAME[a])

	if (filt != "")
		filt = "&& (" filt ")"

	if (noterm == 1)
		noterm = "&& (@.a.term == \"\")"
	else if (noterm == -1)
		noterm = "&& (@.a.term != \"\")"
	else
		noterm = ""

# search for non-terminals (vias) - lists subc vias too
	list = idplist("alloc")
	query("append", list, "(@.type == PSTK) " noterm " " filt)

# take each via found...
	for(;;) {
		obj = idplist("pop", list)
		if (obj == "") break

# ... and check if it needs any change
		needs_action = 0
		for(a = 0; a < fids; a++) {
			if (ACT[a] == 0) continue
			curr = (queryobj(obj, FID[a]) == "" ? remove : add)
			if (curr != ACT[a]) {
				needs_action = 1
				break
			}
		}

		if (!needs_action)
			continue

# If it needs action, a temporary prototype is created (duplicating the
# original prototype) on whcih all changes are done.
# Always operate on a copy, because any change to a live prototype affects
# all padstacks sharing the prototype ID and our target may be only some of
# those padstacks.
		proto = pstkprototmp("dup", obj)

# make all necessary modifications, layer by layer on the temp proto
		for(a = 0; a < fids; a++) {
			if (ACT[a] == 0) continue
			curr = (queryobj(obj, FID[a]) == "" ? remove : add)
			if (curr != ACT[a]) {
				if (ACT[a] == add)
					pstkprotoedit(proto, "copy", FIDNAME[a], FIDSRC[a]);
				else
					pstkprotoedit(proto, "remove", FIDNAME[a]);
			}
			else {
# nothing to do here: requested state matches current state
			}
		}

# insert the resulting prototype at obj's context and change the padstack
# to use it. Note: when the copy is inserted, pcb-rnd automatically does
# a deduplication so this won't result in one proto per padstack.
		pid = pstkprototmp("insert", obj, proto)
		propset(obj, "p/padstack/proto", pid)
		pstkprototmp("free", proto)
	}

# clean up
	for(a = 0; a < fids; a++)
		if (FID[a] != "")
			querycompilefield("free", FID[a])
}

# load ACT[] with operations for the requested side
function setup_act(ACT, req_side, top, bottom, op)
{
	if ((req_side == 0) || (req_side == 1))
		ACT[top] = op
	if ((req_side == 0) || (req_side == 2))
		ACT[bottom] = op
}

# internal action the GUI calls when the button is clicked; this action
# reads dialog box selections to perform the actual job
function via_tent_dialog_cb(    scope,op,side,noterm,filt,ACT)
{
	if (!dad(dlg, "exists", dlg))
		return

	scope = dad(dlg, "get", dlg_wsc);
	op = dad(dlg, "get", dlg_wop);
	side = dad(dlg, "get", dlg_wsd);
	noterm = dad(dlg, "get", dlg_wnt);

	if (scope == 1)
		filt = "@.p.flag.selected"

# load ACT[], the action array depending on requested operation and side
	if (op == 0)       setup_act(ACT, side, top_mask, bottom_mask, remove)
	else if (op == 1)  setup_act(ACT, side, top_mask, bottom_mask, add)
	else if (op == 2)  setup_act(ACT, side, top_paste, bottom_paste, add)
	else if (op == 3)  setup_act(ACT, side, top_paste, bottom_paste, remove)

	if (noterm == 2)
		noterm = -1

	atomic("save")
	atomic("freeze")
	via_tent_(ACT, noterm, filt)
	atomic("unfreeze")
	atomic("block")
	redraw()
}

# action that pops up the GUI dialog
function via_tent_dialog()
{
# never open two instances - if one already exists, return silently
	if (dad(dlg, "exists", dlg))
		return

	dad(dlg, "new")
	dad(dlg, "begin_vbox")
		dad(dlg, "begin_table", 2)
			dad(dlg, "label", "scope");         dlg_wsc = dad(dlg, "enum", "all	selected")
			dad(dlg, "label", "operation");     dlg_wop = dad(dlg, "enum", "tent	untent	paste	unpaste")
			dad(dlg, "label", "side");          dlg_wsd = dad(dlg, "enum", "both	top only	bottom only")
			dad(dlg, "label", "padstack type"); dlg_wnt = dad(dlg, "enum", "vias and terminals	vias only	terminals only")
		dad(dlg, "end")
		dad(dlg, "begin_hbox")
			dad(dlg, "button", "Do it!")
				dad(dlg, "onchange", "via_tent_dialog_cb")
			dad(dlg, "button_closes", "close", 0)
		dad(dlg, "end")
	dad(dlg, "end")
	dad(dlg, "run", "Via tenting and pad pasting", "");
}

# action for the cli and other scripts:
function via_tent(scope, how, layers, noterm    ,ACT,filt)
{
# set up defaults
	if (scope == "") scope = "all"
	if (how == "") how = "tent"
	if (layers == "") layers = "both"
	if (noterm == "") noterm = "none"

# process human CLI arguments
	if (scope == "all") filt = ""
	else if (scope == "selected") filt = "@.p.flag.selected"
	else {
		message("ERROR", "unknown scope: " scope " - needs to be all or selected")
		return
	}

	if (how == "tent")         setup_act(ACT, side, top_mask, bottom_mask, remove)
	else if (how == "untent")  setup_act(ACT, side, top_mask, bottom_mask, add)
	else if (how == "paste")   setup_act(ACT, side, top_paste, bottom_paste, add)
	else if (how == "unpaste") setup_act(ACT, side, top_paste, bottom_paste, remove)
	else {
		message("ERROR", "unknown how: " how " - needs to be one of tent, untent, paste, unpaste")
		return
	}

	if (layers == "both") layers = 0
	else if (layers == "top") layers = 1
	else if (layers == "bottom") layers = 2
	else {
		message("ERROR", "unknown layer: " layer " - needs to be one of both, top or bottom")
		return
	}

	if (noterm == "noterm") noterm = 1;
	else if (noterm == "novia") noterm = -1;
	else if (noterm == "none") noterm = 0;
	else {
		message("ERROR", "unknown noterm: " noter " - needs to be one of noterm, novia or none")
		return
	}

	atomic("save")
	atomic("freeze")
	via_tent_(ACT, noterm, filt)
	atomic("unfreeze")
	atomic("block")
	redraw()
}


# script initialization
BEGIN {
	dlg="via_tent_dialog"
	fgw_func_reg("via_tent_dialog")
	fgw_func_reg("via_tent_dialog_cb")
	fgw_func_reg("via_tent")
}
